;; Copyright (c) 2007, Vladimir Sedach

;; All rights reserved.

;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met: Redistributions of source code must retain the above copyright
;; notice, this list of conditions and the following disclaimer.
;; Redistributions in binary form must reproduce the above copyright
;; notice, this list of conditions and the following disclaimer in the
;; documentation and/or other materials provided with the distribution.

;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

;; Version 1

(declaim (optimize (speed 3)))

(defpackage "VAS-SPARSE-VECTOR"
  (:nicknames "SVEC")
  (:use "COMMON-LISP")
  (:export #:make-sparse-vector
           #:s-ref))

(in-package "VAS-SPARSE-VECTOR")

(defvar *vector-chunk-size* 32)
(declaim (fixnum *vector-chunk-size*))

(defun make-vector-chunk ()
  (make-array (list *vector-chunk-size*) :initial-element nil))

(defun make-sparse-vector ()
  (list (cons 0 (make-vector-chunk))))

(defmacro start-pos (vector-chunk)
  `(car ,vector-chunk))

(defmacro contents (vector-chunk)
  `(cdr ,vector-chunk))

(declaim (inline s-ref (setf s-ref)))

(defun s-ref (v i)
  (declare (fixnum i))
  (loop for lst = v then (cdr lst)
        for start1 of-type fixnum = (the fixnum (start-pos (first lst)))
        for bdry of-type fixnum = (the fixnum (+ start1 (1- *vector-chunk-size*))) do
        (cond ((<= start1 i bdry)
               (return (svref (contents (car lst)) (the fixnum (- i start1)))))
              ((or (null (second lst)) (< bdry i (the fixnum (start-pos (second lst)))))
               (return nil)))))

(defun (setf s-ref) (val v i)
  (declare (fixnum i))
  (let ((chunk-start-pos (- i (mod i *vector-chunk-size*))))
    (loop for lst = v then (cdr lst)
          for start1 of-type fixnum = (start-pos (first lst)) do
          (cond ((< (1- start1) i (+ start1 *vector-chunk-size*))
                 (setf (svref (contents (car lst)) (the fixnum (- i start1))) val)
                 (return))
                ((null (cdr lst)) (let ((new-chunk-contents (make-vector-chunk)))
                                    (setf (svref new-chunk-contents (mod i *vector-chunk-size*)) val
                                          (cdr lst) (list (cons chunk-start-pos new-chunk-contents)))
                                    (return)))
                ((< (the fixnum (+ start1 (1- *vector-chunk-size*))) i (the fixnum (start-pos (second lst))))
                 (let ((new-chunk-contents (make-vector-chunk)))
                   (setf (svref new-chunk-contents (mod i *vector-chunk-size*)) val
                         (cdr lst) (cons (cons chunk-start-pos new-chunk-contents) (cdr lst))))
                 (return)))))
  val)

(defvar *sparse-vector*)
(defvar *regular-vector*)
(defvar *i*)

(defun test-chunks (&optional (number-trials 1000) (max-vector-length 9000))
  (let ((sparse-vector (make-sparse-vector))
        (regular-vector (make-array (list max-vector-length) :initial-element nil)))
    (flet ((check-vectors ()
             (loop for i from 0 below max-vector-length do
                   (unless (equal (aref regular-vector i) (s-ref sparse-vector i))
                     (setf *sparse-vector* sparse-vector
                           *regular-vector* regular-vector
                           *i* i)
                     (break )))))
      (loop repeat number-trials do
            (let ((i (random max-vector-length))
                  (v (random max-vector-length)))
              (setf (aref regular-vector i) v
                    (s-ref sparse-vector i) v)
              (check-vectors))))))